
-------------------------------------------------------------------------------
-- Optimzation
-------------------------------------------------------------------------------

-- GLOBALS: string

local _G = _G
local coroutine = coroutine
local next = next
local SixUI_DisplayGroup_Update = SixUI_DisplayGroup_Update
local table = table
local type = type

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local A, B, C, D = unpack(select(2, ...)) -- AddOn, Border, Constants, Database
local M = A:NewModule("Display Group")
local DisplayFactory = LibStub("SixDisplay-1.0"):CreateFactory({
	glyphs = C.Enum.Glyph,
	glyphFont = C.GLYPH_REGULAR,
	textFont = C.TEXT_REGULAR,
})
local LibDataBroker = LibStub("LibDataBroker-1.1")

-------------------------------------------------------------------------------
-- Module
-------------------------------------------------------------------------------

M.KnownTypes = {
	["data source"] = true,
	["launcher"] = true,
	["progress bar"] = true,
}

function M.SortDisplayConfigData(a, b)
	local aOrder = a.displayConfig.order
	local bOrder = b.displayConfig.order

	if aOrder ~= bOrder then
		return aOrder < bOrder
	else
		return a.displayName < b.displayName
	end
end

function M:OnEnable()
	self:RegisterMessage("SIXDISPLAY_VISIBLITY_CHANGED", "OnVisibilityChanged")
	self:RegisterMessage("SIXUI_CONFIG_UPDATE")
	self:RegisterMessage("SIXUI_PROFILE_UPDATE")
	self:CreateDisplays()
end

function M:OnVisibilityChanged(event, display)
	self:UpdateDisplayGroups(display.config.group)
end

function M:CreateDisplays()
	for displayName, displayConfig in self:IterateDisplayConfigs() do
		if displayConfig.enabled then
			self:CreateDisplay(displayName, displayConfig)
		end
	end
end

function M:CreateDisplay(displayName, displayConfig)
	DisplayFactory:UpsertDisplay(displayName, displayConfig)
end

function M:DeleteDisplays()
	DisplayFactory.pool:ReleaseAll()
end

function M:DeleteDisplay(displayName)
	local display = DisplayFactory:GetDisplay(displayName)

	if display then
		DisplayFactory.pool:Release(display)
	end
end

function M:IterateDisplayConfigs(groupName)
	local displayConfigData = {}

	for displayName, dataObject in LibDataBroker:DataObjectIterator() do
		if type(displayName) == "string" and self.KnownTypes[dataObject.type] then
			table.insert(displayConfigData, {displayName = displayName, dataObject = dataObject, displayConfig = self.config.displayConfigs[dataObject.type][displayName]})
		end
	end

	table.sort(displayConfigData, M.SortDisplayConfigData)

	return coroutine.wrap(function ()
		for _, data in next, displayConfigData do
			if not groupName or data.displayConfig.group == groupName then
				coroutine.yield(data.displayName, data.displayConfig, data.dataObject)
			end
		end
	end)
end

function M:RenanchorDisplays(groupName)
	local anchorDisplay = nil
	local group = _G[groupName]

	for displayName in self:IterateDisplayConfigs(groupName) do
		local display = DisplayFactory:GetDisplay(displayName)

		if display and display:IsShown() then
			display:SetParent(group)
			display:ClearAllPoints()

			if anchorDisplay then
				display:SetPoint("LEFT", anchorDisplay, "RIGHT")
			else
				display:SetPoint("LEFT", group, "LEFT")
			end

			anchorDisplay = display
		end
	end
end

function M:StyleDisplays(groupName)
	local glyphFont, createBorder = C.GLYPH_REGULAR, true

	if groupName == "SixUI_MinimapButtons" then
		glyphFont, createBorder = C.GLYPH_LARGE, false
	end

	for displayName in self:IterateDisplayConfigs(groupName) do
		local display = DisplayFactory:GetDisplay(displayName)

		if display and display:IsShown() then
			if createBorder then
				B:CreateBorder(display)
			else
				B:RemoveBorder(display)
			end

			if display.config.glyphFont ~= glyphFont then
				display:SetConfig({glyphFont = glyphFont})
			end
		end
	end
end

function M:UpdateDisplay(displayName, displayConfig)
	local display = DisplayFactory:GetDisplay(displayName)

	if display then
		display:SetConfig(displayConfig)
	end
end

function M:UpdateDisplayGroups()
	for _, groupName in next, {"SixUI_DisplayGroup1", "SixUI_DisplayGroup2", "SixUI_MinimapButtons", "SixUI_ProgressGroup"} do
		self:UpdateDisplayGroup(groupName)
	end
end

function M:UpdateDisplayGroup(groupName)
	self:RenanchorDisplays(groupName)
	self:StyleDisplays(groupName)
	SixUI_DisplayGroup_Update(_G[groupName])
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

function M:SIXUI_CONFIG_UPDATE(event, key, ...)
	local c, g, k = string.split(".", key)

	if c == "font" then
		self:DeleteDisplays()
		self:CreateDisplays()
		self:UpdateDisplayGroups()
	end
end

function M:SIXUI_PROFILE_UPDATE()
	self:DeleteDisplays()
	self:CreateDisplays()
	self:UpdateDisplayGroups()
end

-------------------------------------------------------------------------------
-- Options
-------------------------------------------------------------------------------

function M:GetAttributeOptions()
	return {
		icon = "Icon",
		glyph = "Glyph",
		label = "Label",
		text = "Text",
		value = "Value",
		suffix = "Suffix",
	}
end

function M:GetGlyphOptions()
	local glyphs = {[""] = "None"}

	for glyph in next, C.Enum.Glyph do
		glyphs[glyph] = glyph
	end

	return glyphs
end

function M:GetGroupOptions()
	return {
		SixUI_DisplayGroup1 = "Chat Frame Bottom",
		SixUI_ProgressGroup = "Chat Frame Right",
		SixUI_DisplayGroup2 = "Minimap Bottom",
		SixUI_MinimapButtons = "Minimap Top",
	}
end

function M:GetGroupSorting()
	return {
		"SixUI_DisplayGroup1",
		"SixUI_ProgressGroup",
		"SixUI_DisplayGroup2",
		"SixUI_MinimapButtons",
	}
end
